package aspects;

import java.util.Stack;

/**
 * 1. (Legal,enterTrusted,_)     -> (Legal,_)
 * 2. (Legal,exitTrusted,_)      -> (Legal,_)
 * 3. (Legal,enterDoPriv,_)      -> (Legal,L)
 * 4. (Legal,exitDoPriv,L)       -> (Legal,_)
 * 5. (Legal,enterUntrusted,_)   -> (Ilegal,L)
 * 6. (Legal,exitDoPriv,*I)       -> (Ilegal,*I)
 * 
 * 7. (Ilegal,enterTrusted,_)    -> (Ilegal,*I)
 * 8. (Ilegal,exitTrusted,*I)     -> (Ilegal,_)
 * 9. (Ilegal,enterDoPriv,I)     -> (Ilegal,I)
 * 10. (Ilegal,exitDoPriv,I)     -> (Ilegal,I)
 * 11. (Ilegal,enterDoPriv,L)    -> (Ilegal,L)
 * 12. (Ilegal,exitDoPriv,L)     -> (Ilegal,L)
 * 13. (Ilegal,enterDoPriv,*I)    -> (Legal,*I)
 * 14. (Ilegal,enterUntrusted,_) -> (Ilegal,I)
 * 15. (Ilegal,exitUntrusted,I)  -> (Ilegal,_)
 * 
 * @author Angel Nunez (angel.nunez@gmail.com)
 *
 */
public class AC_Automaton {

	private final State legal = new Legal();
    private final State ilegal = new Ilegal();
    
    private Stack<StackElem> stack = new Stack<StackElem>();
    private State currentState = legal;
    
    public boolean inLegalState() { return currentState.inLegalState(); }
    public void enterTrusted() { currentState.enterTrusted(); }
    public void exitTrusted() { currentState.exitTrusted(); }
    public void enterUntrusted() { currentState.enterUntrusted(); }
    public void exitUntrusted() { currentState.exitUntrusted(); }
    public void enterDoPriv() { currentState.enterDoPriv(); }
    public void exitDoPriv() { currentState.exitDoPriv(); }
    
    private interface State {
    	boolean inLegalState();
    	void enterTrusted();
    	void exitTrusted();
    	void enterUntrusted();
    	void exitUntrusted();
    	void enterDoPriv();
    	void exitDoPriv();
    }
    
	private class Legal implements State {
		
		public boolean inLegalState() {
			return true;
		}
		
		public void enterTrusted() {
			//trans. 1
		}

		public void exitTrusted() {
			//trans. 2
		}

		public void enterUntrusted() {
			//trans. 5
			stack.push(StackElem.L);
			currentState = ilegal;
		}

		public void exitUntrusted() {
			throw new RuntimeException("Unexpected 'exitUntrusted' in state 'Legal'");
		}

		public void enterDoPriv() {
			//trans. 3
			stack.push(StackElem.L);
		}

		public void exitDoPriv() {
			StackElem peek = stack.peek();
			//trans. 4
			if(peek == StackElem.L) {
				stack.pop();
			}
			//trans. 6
			else if(peek == StackElem.star_I) {
				currentState = ilegal;
			}
			else {
				//because we never get into state legal with stack.peek == I
				throw new RuntimeException("Unexpected 'exitDoPriv' in state 'Legal' with stack.peek == I");
			}
		}
	}
	
	private class Ilegal implements State {
		
		public boolean inLegalState() {
			return false;
		}
		
		public void enterTrusted() {
			//trans. 7
			stack.push(StackElem.star_I);
		}

		public void exitTrusted() {
			//trans. 8
			if(stack.peek() == StackElem.star_I) {
				stack.pop();
			}
			else {
				//because "enterTrusted" always put a 'M' at this state
				throw new RuntimeException("Unexpected 'exitTrusted' in state 'Ilegal' with stack.peek != M");
			}
		}

		public void enterUntrusted() {
			//trans. 14
			stack.push(StackElem.I);
		}

		public void exitUntrusted() {
			//trans. 15
			if(stack.peek() == StackElem.I) {
				stack.pop();
			}
			//trans. 16
			else if(stack.peek() == StackElem.L) {
				stack.pop();
				currentState = legal;
			}
			else {
				//because "enterUntrusted" always put a 'I' at this state
				throw new RuntimeException("Unexpected 'exitUntrusted' in state 'Ilegal' with stack.peek != I");
			}
		}

		public void enterDoPriv() {
			//trans. 9 and 11
			if(stack.peek() == StackElem.I || stack.peek() == StackElem.L) {
				//nothing
			}	
			//trans. 13
			else if(stack.peek() == StackElem.star_I) {
				currentState = legal;
			}
		}

		public void exitDoPriv() {
			//trans. 10 and 12
			if(stack.peek() == StackElem.I || stack.peek() == StackElem.L) {
				//do nothing
			}
			else {
				//because a 'enterDoPriv' with stack.peek == M leads to a transition to state legal
				throw new RuntimeException("Unexpected 'exitDoPriv' in state 'Ilegal' with stack.peek == M");
			}
		}
	}
	
	private enum StackElem {
		L, I, star_I
	}
}
